import asyncio
from typing import AsyncGenerator

import pytest
from decoy import Decoy, matchers

from opentrons.hardware_control.poller import Poller, Reader

POLLING_INTERVAL = 0.1


@pytest.fixture
def mock_reader(decoy: Decoy) -> Reader:
    """Get a mocked out Reader."""
    return decoy.mock(cls=Reader)


@pytest.fixture
async def mock_reader_flow_control(
    decoy: Decoy,
    mock_reader: Reader,
    read_started_event: asyncio.Event,
    ok_to_finish_read_event: asyncio.Event,
) -> None:
    """Configure the Reader with Event flags for controlling read timing."""

    async def _mock_read() -> None:
        read_started_event.set()
        ok_to_finish_read_event.clear()
        await ok_to_finish_read_event.wait()

    decoy.when(await mock_reader.read()).then_do(_mock_read)  # type: ignore[func-returns-value]


@pytest.fixture()
async def read_started_event() -> asyncio.Event:
    """Event to notify a test that a poll has started when using `mock_reader_flow_control`."""
    return asyncio.Event()


@pytest.fixture()
async def ok_to_finish_read_event() -> asyncio.Event:
    """Event to signal a poll to finish when using `mock_reader_flow_control`."""
    return asyncio.Event()


@pytest.fixture
async def subject(
    mock_reader: Reader,
    ok_to_finish_read_event: asyncio.Event,
) -> AsyncGenerator[Poller, None]:
    """Create a poller with a mocked out reader."""
    poller = Poller(reader=mock_reader, interval=POLLING_INTERVAL)
    yield poller
    ok_to_finish_read_event.set()
    await poller.stop()


async def test_poller(decoy: Decoy, mock_reader: Reader, subject: Poller) -> None:
    decoy.verify(await mock_reader.read(), times=0)

    await subject.start()
    decoy.verify(await mock_reader.read(), times=1)

    await subject.wait_next_poll()
    decoy.verify(await mock_reader.read(), times=2)

    await subject.stop()
    decoy.verify(await mock_reader.read(), times=2)


async def test_poller_concurrency(
    mock_reader_flow_control: None,
    read_started_event: asyncio.Event,
    ok_to_finish_read_event: asyncio.Event,
    subject: Poller,
) -> None:
    """It should wait for a full poll before notifying."""
    # wait for the first read to start
    asyncio.create_task(subject.start())
    await asyncio.wait_for(read_started_event.wait(), timeout=4 * subject.interval)

    # subscribe in the middle of the first read
    poll_notification = asyncio.create_task(subject.wait_next_poll())

    # allow the first read to finish, then wait for the second read to start
    read_started_event.clear()
    ok_to_finish_read_event.set()
    await asyncio.wait_for(read_started_event.wait(), timeout=4 * subject.interval)

    # verify that our wait isn't done, because it was kicked off after the first read started
    assert poll_notification.done() is False

    # allow the second read to complete and wait for the third read to start
    read_started_event.clear()
    ok_to_finish_read_event.set()
    await asyncio.wait_for(read_started_event.wait(), timeout=4 * subject.interval)

    # verify the waiter has now been notified since it's been through the full second read
    assert poll_notification.done() is True


async def test_poller_stop_waits_for_poll(
    mock_reader_flow_control: None,
    read_started_event: asyncio.Event,
    ok_to_finish_read_event: asyncio.Event,
    subject: Poller,
) -> None:
    # wait for the first read to start
    asyncio.create_task(subject.start())
    await asyncio.wait_for(read_started_event.wait(), timeout=4 * subject.interval)

    # request a stop in the middle of the first read
    stop_request = asyncio.create_task(subject.stop())
    await asyncio.sleep(4 * subject.interval)
    assert stop_request.done() is False

    ok_to_finish_read_event.set()
    await asyncio.sleep(2 * subject.interval)
    assert stop_request.done() is True


async def test_poller_start_error(
    decoy: Decoy, mock_reader: Reader, subject: Poller
) -> None:
    """It should raise in start if read errors."""
    decoy.when(await mock_reader.read()).then_raise(RuntimeError("oh no"))  # type: ignore[func-returns-value]

    with pytest.raises(RuntimeError, match="oh no"):
        await subject.start()

    decoy.verify(
        mock_reader.on_error(matchers.ErrorMatching(RuntimeError, match="oh no")),
        times=1,
    )


async def test_poller_wait_next_poll_error(
    decoy: Decoy, mock_reader: Reader, subject: Poller
) -> None:
    """It should raise in wait_next_poll if read errors."""
    await subject.start()

    decoy.when(await mock_reader.read()).then_raise(RuntimeError("oh no"))  # type: ignore[func-returns-value]

    with pytest.raises(RuntimeError, match="oh no"):
        await subject.wait_next_poll()

    decoy.verify(
        mock_reader.on_error(matchers.ErrorMatching(RuntimeError, match="oh no")),
        times=1,
    )


async def test_poller_cancelled_wait(
    mock_reader_flow_control: None,
    subject: Poller,
    read_started_event: asyncio.Event,
    ok_to_finish_read_event: asyncio.Event,
) -> None:
    """If one waiter is cancelled, other waiters should still work."""
    asyncio.create_task(subject.start())
    await read_started_event.wait()
    read_started_event.clear()

    # Before letting the read complete, set up new wait tasks
    wait_task_1 = asyncio.create_task(subject.wait_next_poll())
    wait_task_2 = asyncio.create_task(subject.wait_next_poll())
    ok_to_finish_read_event.set()

    await read_started_event.wait()
    read_started_event.clear()
    wait_task_1.cancel()
    ok_to_finish_read_event.set()

    assert (
        subject._poll_forever_task is not None and not subject._poll_forever_task.done()
    )

    await asyncio.sleep(2 * subject.interval)
    assert wait_task_2.done() is True



"""Test the stacker coommon code."""

from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import CommandPreconditionViolated
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.flex_stacker import common as subject
from opentrons.protocol_engine.execution.equipment import (
    EquipmentHandler,
    LoadedLabwarePoolData,
)
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    BatchLabwareLocationUpdate,
    BatchLoadedLabwareUpdate,
    FlexStackerStateUpdate,
    LabwareLidUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    InStackerHopperLocation,
    LabwareLocation,
    LabwareLocationSequence,
    LabwareOffset,
    LabwareUri,
    LoadedLabware,
    ModuleLocation,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
    OnLabwareOffsetLocationSequenceComponent,
    StackerStoredLabwareGroup,
)


@pytest.mark.parametrize(
    "group_to_check,result",
    [
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId=None,
                lidLabwareId=None,
            ),
            subject.GroupWithLocationSequences(
                primary=subject.LabwareWithLocationSequence(
                    labwareId="some-primary", locationSequence=[sentinel.base_location]
                ),
                adapter=None,
                lid=None,
            ),
            id="primary-only",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId="some-adapter",
                lidLabwareId=None,
            ),
            subject.GroupWithLocationSequences(
                primary=subject.LabwareWithLocationSequence(
                    labwareId="some-primary",
                    locationSequence=[
                        OnLabwareLocationSequenceComponent(
                            labwareId="some-adapter", lidId=None
                        ),
                        sentinel.base_location,
                    ],
                ),
                adapter=subject.LabwareWithLocationSequence(
                    labwareId="some-adapter", locationSequence=[sentinel.base_location]
                ),
                lid=None,
            ),
            id="primary-and-adapter",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId=None,
                lidLabwareId="some-lid",
            ),
            subject.GroupWithLocationSequences(
                primary=subject.LabwareWithLocationSequence(
                    labwareId="some-primary",
                    locationSequence=[
                        sentinel.base_location,
                    ],
                ),
                adapter=None,
                lid=subject.LabwareWithLocationSequence(
                    labwareId="some-lid",
                    locationSequence=[
                        OnLabwareLocationSequenceComponent(
                            labwareId="some-primary", lidId="some-lid"
                        ),
                        sentinel.base_location,
                    ],
                ),
            ),
            id="primary-and-lid",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId="some-adapter",
                lidLabwareId="some-lid",
            ),
            subject.GroupWithLocationSequences(
                primary=subject.LabwareWithLocationSequence(
                    labwareId="some-primary",
                    locationSequence=[
                        OnLabwareLocationSequenceComponent(
                            labwareId="some-adapter", lidId=None
                        ),
                        sentinel.base_location,
                    ],
                ),
                adapter=subject.LabwareWithLocationSequence(
                    labwareId="some-adapter", locationSequence=[sentinel.base_location]
                ),
                lid=subject.LabwareWithLocationSequence(
                    labwareId="some-lid",
                    locationSequence=[
                        OnLabwareLocationSequenceComponent(
                            labwareId="some-primary", lidId="some-lid"
                        ),
                        OnLabwareLocationSequenceComponent(
                            labwareId="some-adapter", lidId=None
                        ),
                        sentinel.base_location,
                    ],
                ),
            ),
            id="primary-lid-and-adapter",
        ),
    ],
)
def test_labware_locations_for_group(
    group_to_check: StackerStoredLabwareGroup,
    result: subject.GroupWithLocationSequences,
) -> None:
    """It should bind locations to groups and fall back to None for missing groups."""
    assert (
        subject.labware_locations_for_group(group_to_check, [sentinel.base_location])
        == result
    )


@pytest.mark.parametrize(
    "group,result,known",
    [
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId=None,
                lidLabwareId=None,
            ),
            [sentinel.primary_base],
            True,
            id="no-adapter-yes-known",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId="some-adapter",
                lidLabwareId=None,
            ),
            [sentinel.adapter_base],
            True,
            id="yes-adapter-yes-known",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId=None,
                lidLabwareId=None,
            ),
            [sentinel.default],
            False,
            id="no-adapter-no-known",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="some-primary",
                adapterLabwareId="some-adapter",
                lidLabwareId=None,
            ),
            [sentinel.default],
            False,
            id="yes-adapter-no-known",
        ),
    ],
)
def test_labware_location_base_sequence(
    group: StackerStoredLabwareGroup,
    result: LabwareLocationSequence,
    known: bool,
    state_view: StateView,
    decoy: Decoy,
) -> None:
    """It should query the correct labware's location."""
    decoy.when(state_view.labware.known("some-primary")).then_return(known)
    decoy.when(state_view.labware.known("some-adapter")).then_return(known)
    decoy.when(state_view.geometry.get_location_sequence("some-primary")).then_return(
        [sentinel.primary_base]
    )
    decoy.when(state_view.geometry.get_location_sequence("some-adapter")).then_return(
        [sentinel.adapter_base]
    )
    assert (
        subject.labware_location_base_sequence(group, state_view, [sentinel.default])
        == result
    )


@pytest.mark.parametrize(
    "groups,result",
    [
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-1", locationSequence=[sentinel.first]
                    ),
                    adapter=None,
                    lid=None,
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-2", locationSequence=[sentinel.second]
                    ),
                    adapter=None,
                    lid=None,
                ),
            ],
            [[sentinel.first], [sentinel.second]],
            id="not-empty",
        ),
        pytest.param([], [], id="empty"),
    ],
)
def test_primary_location_sequences(
    groups: list[subject.GroupWithLocationSequences],
    result: list[LabwareLocationSequence],
) -> None:
    """It should handle empty and non-empty lists."""
    assert subject.primary_location_sequences(groups) == result


@pytest.mark.parametrize(
    "groups,result",
    [
        pytest.param([], [], id="empty-list"),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-1",
                        locationSequence=[sentinel.first_primary],
                    ),
                    adapter=None,
                    lid=None,
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-2",
                        locationSequence=[sentinel.second_primary],
                    ),
                    adapter=None,
                    lid=None,
                ),
            ],
            None,
            id="no-adapter",
        ),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-1",
                        locationSequence=[sentinel.first_primary],
                    ),
                    adapter=subject.LabwareWithLocationSequence(
                        labwareId="adapter-id-1",
                        locationSequence=[sentinel.first_adapter],
                    ),
                    lid=None,
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-2",
                        locationSequence=[sentinel.second_primary],
                    ),
                    adapter=subject.LabwareWithLocationSequence(
                        labwareId="adapter-id-2",
                        locationSequence=[sentinel.second_adapter],
                    ),
                    lid=None,
                ),
            ],
            [[sentinel.first_adapter], [sentinel.second_adapter]],
            id="adapter",
        ),
    ],
)
def test_adapter_location_sequences(
    groups: list[subject.GroupWithLocationSequences],
    result: list[LabwareLocationSequence] | None,
) -> None:
    """It should handle lists, empty lists, and no-adapter cases."""
    assert subject.adapter_location_sequences(groups) == result


@pytest.mark.parametrize(
    "groups,result,has_adapter",
    [
        pytest.param([], [], True, id="yes-adapter-empty"),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-1",
                        locationSequence=[sentinel.primary_location_1],
                    ),
                    adapter=subject.LabwareWithLocationSequence(
                        labwareId="adapter-1",
                        locationSequence=[sentinel.adapter_location_1],
                    ),
                    lid=None,
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-2",
                        locationSequence=[sentinel.primary_location_2],
                    ),
                    adapter=subject.LabwareWithLocationSequence(
                        labwareId="adapter-2",
                        locationSequence=[sentinel.adapter_location_2],
                    ),
                    lid=None,
                ),
            ],
            [[sentinel.adapter_location_1], [sentinel.adapter_location_2]],
            True,
            id="yes-adapter",
        ),
        pytest.param([], None, False, id="no-adapter-empty"),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-1",
                        locationSequence=[sentinel.primary_location_1],
                    ),
                    adapter=None,
                    lid=None,
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-2",
                        locationSequence=[sentinel.primary_location_2],
                    ),
                    adapter=None,
                    lid=None,
                ),
            ],
            None,
            False,
            id="no-adapter",
        ),
    ],
)
def test_adapter_location_sequences_with_default(
    groups: list[subject.GroupWithLocationSequences],
    result: list[LabwareLocationSequence] | None,
    has_adapter: bool,
) -> None:
    """It should handle cases where there is no adapter."""
    assert (
        subject.adapter_location_sequences_with_default(
            groups, sentinel.adapter_def if has_adapter else None
        )
        == result
    )


@pytest.mark.parametrize(
    "groups,result",
    [
        pytest.param([], [], id="empty-list"),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-1",
                        locationSequence=[sentinel.first_primary],
                    ),
                    adapter=None,
                    lid=None,
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-2",
                        locationSequence=[sentinel.second_primary],
                    ),
                    adapter=None,
                    lid=None,
                ),
            ],
            None,
            id="no-lid",
        ),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-1",
                        locationSequence=[sentinel.first_primary],
                    ),
                    adapter=None,
                    lid=subject.LabwareWithLocationSequence(
                        labwareId="lid-id-1",
                        locationSequence=[sentinel.first_lid],
                    ),
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-id-2",
                        locationSequence=[sentinel.second_primary],
                    ),
                    adapter=None,
                    lid=subject.LabwareWithLocationSequence(
                        labwareId="lid-id-2",
                        locationSequence=[sentinel.second_lid],
                    ),
                ),
            ],
            [[sentinel.first_lid], [sentinel.second_lid]],
            id="lid",
        ),
    ],
)
def test_lid_location_sequences(
    groups: list[subject.GroupWithLocationSequences],
    result: list[LabwareLocationSequence] | None,
) -> None:
    """It should handle lists, empty lists, and no-adapter cases."""
    assert subject.lid_location_sequences(groups) == result


@pytest.mark.parametrize(
    "groups,result,has_lid",
    [
        pytest.param([], [], True, id="yes-lid-empty"),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-1",
                        locationSequence=[sentinel.primary_location_1],
                    ),
                    adapter=None,
                    lid=subject.LabwareWithLocationSequence(
                        labwareId="lid-1",
                        locationSequence=[sentinel.lid_location_1],
                    ),
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-2",
                        locationSequence=[sentinel.primary_location_2],
                    ),
                    adapter=None,
                    lid=subject.LabwareWithLocationSequence(
                        labwareId="lid-2",
                        locationSequence=[sentinel.lid_location_2],
                    ),
                ),
            ],
            [[sentinel.lid_location_1], [sentinel.lid_location_2]],
            True,
            id="yes-lid",
        ),
        pytest.param([], None, False, id="no-lid-empty"),
        pytest.param(
            [
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-1",
                        locationSequence=[sentinel.primary_location_1],
                    ),
                    adapter=None,
                    lid=None,
                ),
                subject.GroupWithLocationSequences(
                    primary=subject.LabwareWithLocationSequence(
                        labwareId="primary-2",
                        locationSequence=[sentinel.primary_location_2],
                    ),
                    adapter=None,
                    lid=None,
                ),
            ],
            None,
            False,
            id="no-lid",
        ),
    ],
)
def test_lid_location_sequences_with_default(
    groups: list[subject.GroupWithLocationSequences],
    result: list[LabwareLocationSequence] | None,
    has_lid: bool,
) -> None:
    """It should handle cases where there is no adapter."""
    assert (
        subject.lid_location_sequences_with_default(
            groups, sentinel.lid_def if has_lid else None
        )
        == result
    )


@pytest.mark.parametrize("known", [True, False])
@pytest.mark.parametrize(
    "id_groups",
    [
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-1",
                    lidLabwareId=None,
                    adapterLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-2",
                    lidLabwareId=None,
                    adapterLabwareId=None,
                ),
            ],
            id="primaries-only",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-1",
                    lidLabwareId="lid-1",
                    adapterLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-2",
                    lidLabwareId="lid-2",
                    adapterLabwareId=None,
                ),
            ],
            id="primaries-and-lids",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-1",
                    lidLabwareId=None,
                    adapterLabwareId="adapter-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-2",
                    lidLabwareId=None,
                    adapterLabwareId="adapter-2",
                ),
            ],
            id="primaries-and-adapters",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-1",
                    lidLabwareId="lid-1",
                    adapterLabwareId="adapter-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-2",
                    lidLabwareId="lid-2",
                    adapterLabwareId="adapter-2",
                ),
            ],
            id="primaries-adapters-and-lids",
        ),
        pytest.param([], id="empty-list"),
    ],
)
def test_check_if_labware_preloaded(
    state_view: StateView,
    decoy: Decoy,
    id_groups: list[StackerStoredLabwareGroup],
    known: bool,
) -> None:
    """It should handle stack combinations."""
    if len(id_groups) == 0 and known:
        pytest.skip()
    for group in id_groups:
        decoy.when(state_view.labware.known(group.primaryLabwareId)).then_return(known)
        if group.adapterLabwareId:
            decoy.when(state_view.labware.known(group.adapterLabwareId)).then_return(
                known
            )
        if group.lidLabwareId:
            decoy.when(state_view.labware.known(group.lidLabwareId)).then_return(known)
    assert (
        subject.check_if_labware_preloaded(
            ids=id_groups,
            state_view=state_view,
        )
        is known
    )


def test_check_preloaded_labware_primary_consistency(
    state_view: StateView,
    decoy: Decoy,
) -> None:
    """It should check definitions are consistent between the pool and the loaded labware."""
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.primary_definition)
    ).then_return(LabwareUri("primary-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.adapter_definition)
    ).then_return(LabwareUri("adapter-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.lid_definition)
    ).then_return(LabwareUri("lid-uri"))

    decoy.when(state_view.labware.get_definition_uri("primary-id-1")).then_return(
        LabwareUri("other-uri")
    )

    with pytest.raises(
        CommandPreconditionViolated,
        match="previous labware groups specify primary URI.*and this one specifies.*",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            state_view,
        )

    decoy.when(state_view.labware.get_definition_uri("primary-id-1")).then_return(
        LabwareUri("primary-uri")
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="but previous labware groups specify no adapter and this one does",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId=None,
                )
            ],
            state_view,
        )
    decoy.when(state_view.labware.get_location("primary-id-1")).then_return(
        ModuleLocation(moduleId="some-id")
    )

    with pytest.raises(
        CommandPreconditionViolated,
        match="All existing labware.*must be currently OFF_DECK",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            state_view,
        )
        decoy.when(state_view.labware.get_location("primary-id-1")).then_return(
            OFF_DECK_LOCATION
        )
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            state_view,
        )


def test_check_preloaded_labware_adapter_consistency(
    state_view: StateView,
    decoy: Decoy,
) -> None:
    """It should check definitions are consistent between the pool and the loaded labware."""
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.primary_definition)
    ).then_return(LabwareUri("primary-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.adapter_definition)
    ).then_return(LabwareUri("adapter-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.lid_definition)
    ).then_return(LabwareUri("lid-uri"))

    decoy.when(state_view.labware.get_definition_uri("primary-id-1")).then_return(
        LabwareUri("primary-uri")
    )
    decoy.when(state_view.labware.get_definition_uri("primary-id-2")).then_return(
        LabwareUri("primary-uri")
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="previous labware groups specify an adapter and this one does not",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            sentinel.adapter_definition,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            state_view,
        )
    decoy.when(state_view.labware.get_definition_uri("adapter-id-1")).then_return(
        LabwareUri("other-uri")
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="previous labware groups specify adapter URI.*and this one specifies.*",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            sentinel.adapter_definition,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId=None,
                )
            ],
            state_view,
        )
    decoy.when(state_view.labware.get_location("adapter-id-1")).then_return(
        ModuleLocation(moduleId="some-module")
    )
    decoy.when(state_view.labware.get_definition_uri("adapter-id-1")).then_return(
        LabwareUri("adapter-uri")
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="All existing adapters.*must be currently OFF_DECK",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            sentinel.adapter_definition,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId=None,
                )
            ],
            state_view,
        )
    decoy.when(state_view.labware.get_location("adapter-id-1")).then_return(
        OFF_DECK_LOCATION
    )
    decoy.when(state_view.labware.get_location("primary-id-1")).then_return(
        OFF_DECK_LOCATION
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="Existing labware groups.*must already be associated",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            sentinel.adapter_definition,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId=None,
                )
            ],
            state_view,
        )
    decoy.when(state_view.labware.get_location("primary-id-1")).then_return(
        OnLabwareLocation(labwareId="adapter-id-1")
    )
    subject.check_preloaded_labware(
        sentinel.primary_definition,
        sentinel.adapter_definition,
        None,
        [
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-id-1",
                adapterLabwareId="adapter-id-1",
                lidLabwareId=None,
            )
        ],
        state_view,
    )


def test_check_preloaded_labware_lid_consistency(
    state_view: StateView, decoy: Decoy
) -> None:
    """It should check for consistency of preloaded lids."""
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.primary_definition)
    ).then_return(LabwareUri("primary-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.lid_definition)
    ).then_return(LabwareUri("lid-uri"))

    decoy.when(state_view.labware.get_definition_uri("primary-id-1")).then_return(
        LabwareUri("primary-uri")
    )
    decoy.when(state_view.labware.get_location("primary-id-1")).then_return(
        OFF_DECK_LOCATION
    )

    with pytest.raises(
        CommandPreconditionViolated,
        match="previous labware groups specify a lid and this one does not",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            sentinel.lid_definition,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            state_view,
        )

    with pytest.raises(
        CommandPreconditionViolated,
        match="previous labware groups did not specify a lid and this one does",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-1",
                )
            ],
            state_view,
        )

    decoy.when(state_view.labware.get_definition_uri("lid-id-1")).then_return(
        LabwareUri("other-uri")
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="previous labware groups specify lid URI.*and this one specifies.*",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            sentinel.lid_definition,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-1",
                )
            ],
            state_view,
        )

    decoy.when(state_view.labware.get_definition_uri("lid-id-1")).then_return(
        LabwareUri("lid-uri")
    )
    decoy.when(state_view.labware.get_location("lid-id-1")).then_return(
        OFF_DECK_LOCATION
    )
    decoy.when(state_view.labware.get_lid_id_by_labware_id("primary-id-1")).then_return(
        "lid-id-1"
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="Existing labware groups.*must already be associated",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            sentinel.lid_definition,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-1",
                )
            ],
            state_view,
        )

    decoy.when(state_view.labware.get_location("lid-id-1")).then_return(
        OnLabwareLocation(labwareId="primary-id-1")
    )

    decoy.when(state_view.labware.get_lid_id_by_labware_id("primary-id-1")).then_return(
        None
    )
    with pytest.raises(
        CommandPreconditionViolated,
        match="Existing labware groups.*must already be associated",
    ):
        subject.check_preloaded_labware(
            sentinel.primary_definition,
            None,
            sentinel.lid_definition,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-1",
                )
            ],
            state_view,
        )

    decoy.when(state_view.labware.get_location("lid-id-1")).then_return(
        OnLabwareLocation(labwareId="primary-id-1")
    )
    decoy.when(state_view.labware.get_lid_id_by_labware_id("primary-id-1")).then_return(
        "lid-id-1"
    )
    subject.check_preloaded_labware(
        sentinel.primary_definition,
        None,
        sentinel.lid_definition,
        [
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-id-1",
                adapterLabwareId=None,
                lidLabwareId="lid-id-1",
            )
        ],
        state_view,
    )


async def test_assign_n_labware_happypath(state_view: StateView, decoy: Decoy) -> None:
    """When all conditions are met, it should assign labware using state updates."""
    decoy.when(state_view.labware.known("primary-1")).then_return(True)
    decoy.when(state_view.labware.known("primary-2")).then_return(True)
    decoy.when(state_view.labware.known("adapter-1")).then_return(True)
    decoy.when(state_view.labware.known("adapter-2")).then_return(True)
    decoy.when(state_view.labware.known("lid-1")).then_return(True)
    decoy.when(state_view.labware.known("lid-2")).then_return(True)
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.primary_definition)
    ).then_return(LabwareUri("primary-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.adapter_definition)
    ).then_return(LabwareUri("adapter-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.lid_definition)
    ).then_return(LabwareUri("lid-uri"))
    decoy.when(state_view.labware.get_definition_uri("primary-1")).then_return(
        LabwareUri("primary-uri")
    )
    decoy.when(state_view.labware.get_definition_uri("primary-2")).then_return(
        LabwareUri("primary-uri")
    )
    decoy.when(state_view.labware.get_definition_uri("adapter-1")).then_return(
        LabwareUri("adapter-uri")
    )
    decoy.when(state_view.labware.get_definition_uri("adapter-2")).then_return(
        LabwareUri("adapter-uri")
    )
    decoy.when(state_view.labware.get_definition_uri("lid-1")).then_return(
        LabwareUri("lid-uri")
    )
    decoy.when(state_view.labware.get_definition_uri("lid-2")).then_return(
        LabwareUri("lid-uri")
    )
    decoy.when(state_view.labware.get_location("primary-1")).then_return(
        OnLabwareLocation(labwareId="adapter-1")
    )
    decoy.when(state_view.labware.get_location("primary-2")).then_return(
        OnLabwareLocation(labwareId="adapter-2")
    )
    decoy.when(state_view.labware.get_location("adapter-1")).then_return(
        OFF_DECK_LOCATION
    )
    decoy.when(state_view.labware.get_location("adapter-2")).then_return(
        OFF_DECK_LOCATION
    )
    decoy.when(state_view.labware.get_location("lid-1")).then_return(
        OnLabwareLocation(labwareId="primary-1")
    )
    decoy.when(state_view.labware.get_location("lid-2")).then_return(
        OnLabwareLocation(labwareId="primary-2")
    )
    decoy.when(state_view.labware.get_lid_id_by_labware_id("primary-1")).then_return(
        "lid-1"
    )
    decoy.when(state_view.labware.get_lid_id_by_labware_id("primary-2")).then_return(
        "lid-2"
    )
    state, stored = await subject.assign_n_labware(
        sentinel.primary_definition,
        sentinel.adapter_definition,
        sentinel.lid_definition,
        "module-id",
        [
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-1",
                adapterLabwareId="adapter-1",
                lidLabwareId="lid-1",
            ),
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-2",
                adapterLabwareId="adapter-2",
                lidLabwareId="lid-2",
            ),
        ],
        [
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-pre-1",
                adapterLabwareId="adapter-pre-1",
                lidLabwareId="lid-pre-1",
            ),
        ],
        state_view,
    )

    assert state == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id="module-id",
            contained_labware_bottom_first=[
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-1",
                    adapterLabwareId="adapter-pre-1",
                    lidLabwareId="lid-pre-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-1",
                    adapterLabwareId="adapter-1",
                    lidLabwareId="lid-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-2",
                    adapterLabwareId="adapter-2",
                    lidLabwareId="lid-2",
                ),
            ],
        ),
        batch_labware_location=BatchLabwareLocationUpdate(
            new_locations_by_id={
                "adapter-1": InStackerHopperLocation(moduleId="module-id"),
                "adapter-2": InStackerHopperLocation(moduleId="module-id"),
            },
            new_offset_ids_by_id={
                "adapter-1": None,
                "adapter-2": None,
                "primary-1": None,
                "primary-2": None,
                "lid-1": None,
                "lid-2": None,
            },
        ),
    )

    assert stored == [
        StackerStoredLabwareGroup(
            primaryLabwareId="primary-pre-1",
            adapterLabwareId="adapter-pre-1",
            lidLabwareId="lid-pre-1",
        ),
        StackerStoredLabwareGroup(
            primaryLabwareId="primary-1",
            adapterLabwareId="adapter-1",
            lidLabwareId="lid-1",
        ),
        StackerStoredLabwareGroup(
            primaryLabwareId="primary-2",
            adapterLabwareId="adapter-2",
            lidLabwareId="lid-2",
        ),
    ]


async def test_assign_n_labware_no_offsets_for_unspecified(
    state_view: StateView, decoy: Decoy
) -> None:
    """When all conditions are met, it should assign labware using state updates."""
    decoy.when(state_view.labware.known("primary-1")).then_return(True)
    decoy.when(state_view.labware.known("primary-2")).then_return(True)
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.primary_definition)
    ).then_return(LabwareUri("primary-uri"))
    decoy.when(state_view.labware.get_definition_uri("primary-1")).then_return(
        LabwareUri("primary-uri")
    )
    decoy.when(state_view.labware.get_definition_uri("primary-2")).then_return(
        LabwareUri("primary-uri")
    )
    decoy.when(state_view.labware.get_location("primary-1")).then_return(
        OFF_DECK_LOCATION
    )
    decoy.when(state_view.labware.get_location("primary-2")).then_return(
        OFF_DECK_LOCATION
    )
    state, stored = await subject.assign_n_labware(
        sentinel.primary_definition,
        None,
        None,
        "module-id",
        [
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-1",
                adapterLabwareId=None,
                lidLabwareId=None,
            ),
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-2",
                adapterLabwareId=None,
                lidLabwareId=None,
            ),
        ],
        [
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-pre-1",
                adapterLabwareId=None,
                lidLabwareId=None,
            ),
        ],
        state_view,
    )

    assert state == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id="module-id",
            contained_labware_bottom_first=[
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
        ),
        batch_labware_location=BatchLabwareLocationUpdate(
            new_locations_by_id={
                "primary-1": InStackerHopperLocation(moduleId="module-id"),
                "primary-2": InStackerHopperLocation(moduleId="module-id"),
            },
            new_offset_ids_by_id={
                "primary-1": None,
                "primary-2": None,
            },
        ),
    )

    assert stored == [
        StackerStoredLabwareGroup(
            primaryLabwareId="primary-pre-1",
            adapterLabwareId=None,
            lidLabwareId=None,
        ),
        StackerStoredLabwareGroup(
            primaryLabwareId="primary-1",
            adapterLabwareId=None,
            lidLabwareId=None,
        ),
        StackerStoredLabwareGroup(
            primaryLabwareId="primary-2",
            adapterLabwareId=None,
            lidLabwareId=None,
        ),
    ]


@pytest.mark.parametrize(
    "ids,current_contains",
    [
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            id="has-contents",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            [],
            id="no-contents",
        ),
    ],
)
async def test_build_n_labware_happypath_primary_only(
    ids: list[StackerStoredLabwareGroup],
    current_contains: list[StackerStoredLabwareGroup],
    equipment: EquipmentHandler,
    decoy: Decoy,
) -> None:
    """It should build the specified labware."""
    for id_group in ids:
        decoy.when(
            await equipment.load_labware_pool_from_definitions(
                pool_primary_definition=sentinel.primary_labware_def,
                pool_adapter_definition=None,
                pool_lid_definition=None,
                location=InStackerHopperLocation(moduleId="module-id"),
                primary_id=id_group.primaryLabwareId,
                adapter_id=None,
                lid_id=None,
            )
        ).then_return(
            LoadedLabwarePoolData(
                primary_labware=LoadedLabware(
                    id=id_group.primaryLabwareId,
                    loadName="some-load-name",
                    definitionUri="some-uri",
                    location=InStackerHopperLocation(moduleId="module-id"),
                    lid_id=None,
                    offsetId=None,
                    displayName=None,
                ),
                adapter_labware=None,
                lid_labware=None,
            )
        )
    state, stored = await subject.build_n_labware_with_ids(
        pool_primary_definition=sentinel.primary_labware_def,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        module_id="module-id",
        ids=ids,
        current_contained_labware=current_contains,
        equipment=equipment,
    )
    assert stored == current_contains + ids
    assert state == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id="module-id",
            contained_labware_bottom_first=current_contains + ids,
        ),
        batch_loaded_labware=BatchLoadedLabwareUpdate(
            new_locations_by_id={
                group.primaryLabwareId: InStackerHopperLocation(moduleId="module-id")
                for group in ids
            },
            offset_ids_by_id={group.primaryLabwareId: None for group in ids},
            display_names_by_id={group.primaryLabwareId: None for group in ids},
            definitions_by_id={
                group.primaryLabwareId: sentinel.primary_labware_def for group in ids
            },
        ),
        labware_lid=LabwareLidUpdate(parent_labware_ids=[], lid_ids=[]),
    )


@pytest.mark.parametrize(
    "ids,current_contains",
    [
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId="adapter-id-2",
                    lidLabwareId=None,
                ),
            ],
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-1",
                    adapterLabwareId="adapter-pre-1",
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-2",
                    adapterLabwareId="adapter-pre-2",
                    lidLabwareId=None,
                ),
            ],
            id="has-contents",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId="adapter-id-2",
                    lidLabwareId=None,
                ),
            ],
            [],
            id="no-contents",
        ),
    ],
)
async def test_build_n_labware_happypath_primary_and_adapter(
    ids: list[StackerStoredLabwareGroup],
    current_contains: list[StackerStoredLabwareGroup],
    equipment: EquipmentHandler,
    decoy: Decoy,
) -> None:
    """It should build the specified labware."""
    for id_group in ids:
        assert id_group.adapterLabwareId
        decoy.when(
            await equipment.load_labware_pool_from_definitions(
                pool_primary_definition=sentinel.primary_labware_def,
                pool_adapter_definition=sentinel.adapter_labware_def,
                pool_lid_definition=None,
                location=InStackerHopperLocation(moduleId="module-id"),
                primary_id=id_group.primaryLabwareId,
                adapter_id=id_group.adapterLabwareId,
                lid_id=None,
            )
        ).then_return(
            LoadedLabwarePoolData(
                primary_labware=LoadedLabware(
                    id=id_group.primaryLabwareId,
                    loadName="some-load-name",
                    definitionUri="some-uri",
                    location=OnLabwareLocation(labwareId=id_group.adapterLabwareId),
                    lid_id=None,
                    offsetId=None,
                    displayName=None,
                ),
                adapter_labware=LoadedLabware(
                    id=id_group.adapterLabwareId,
                    loadName="adapter-load-name",
                    definitionUri="adapter-uri",
                    location=InStackerHopperLocation(
                        moduleId="module-id",
                    ),
                    lid_id=None,
                    offsetId=None,
                    displayName=None,
                ),
                lid_labware=None,
            )
        )
    state, stored = await subject.build_n_labware_with_ids(
        pool_primary_definition=sentinel.primary_labware_def,
        pool_adapter_definition=sentinel.adapter_labware_def,
        pool_lid_definition=None,
        module_id="module-id",
        ids=ids,
        current_contained_labware=current_contains,
        equipment=equipment,
    )
    assert stored == current_contains + ids
    location_ids: dict[str, LabwareLocation] = {}
    ids_flat: list[str] = []
    defs: dict[str, LabwareDefinition] = {}
    for group in ids:
        assert group.adapterLabwareId
        location_ids[group.primaryLabwareId] = OnLabwareLocation(
            labwareId=group.adapterLabwareId
        )
        location_ids[group.adapterLabwareId] = InStackerHopperLocation(
            moduleId="module-id"
        )
        ids_flat.extend([group.primaryLabwareId, group.adapterLabwareId])
        defs[group.primaryLabwareId] = sentinel.primary_labware_def
        defs[group.adapterLabwareId] = sentinel.adapter_labware_def

    assert state == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id="module-id",
            contained_labware_bottom_first=current_contains + ids,
        ),
        batch_loaded_labware=BatchLoadedLabwareUpdate(
            new_locations_by_id=location_ids,
            offset_ids_by_id={id: None for id in ids_flat},
            display_names_by_id={id: None for id in ids_flat},
            definitions_by_id=defs,
        ),
        labware_lid=LabwareLidUpdate(parent_labware_ids=[], lid_ids=[]),
    )


@pytest.mark.parametrize(
    "ids,current_contains",
    [
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-2",
                ),
            ],
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-pre-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-2",
                    adapterLabwareId=None,
                    lidLabwareId="lid-pre-2",
                ),
            ],
            id="has-contents",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-2",
                ),
            ],
            [],
            id="no-contents",
        ),
    ],
)
async def test_build_n_labware_happypath_primary_and_lid(
    ids: list[StackerStoredLabwareGroup],
    current_contains: list[StackerStoredLabwareGroup],
    equipment: EquipmentHandler,
    decoy: Decoy,
) -> None:
    """It should build the specified labware."""
    for id_group in ids:
        assert id_group.lidLabwareId
        decoy.when(
            await equipment.load_labware_pool_from_definitions(
                pool_primary_definition=sentinel.primary_labware_def,
                pool_adapter_definition=None,
                pool_lid_definition=sentinel.lid_labware_def,
                location=InStackerHopperLocation(moduleId="module-id"),
                primary_id=id_group.primaryLabwareId,
                adapter_id=None,
                lid_id=id_group.lidLabwareId,
            )
        ).then_return(
            LoadedLabwarePoolData(
                primary_labware=LoadedLabware(
                    id=id_group.primaryLabwareId,
                    loadName="some-load-name",
                    definitionUri="some-uri",
                    location=InStackerHopperLocation(
                        moduleId="module-id",
                    ),
                    lid_id=id_group.lidLabwareId,
                    offsetId=None,
                    displayName=None,
                ),
                adapter_labware=None,
                lid_labware=LoadedLabware(
                    id=id_group.lidLabwareId,
                    loadName="lid-load-name",
                    definitionUri="lid-uri",
                    location=OnLabwareLocation(
                        labwareId=id_group.primaryLabwareId,
                    ),
                    lid_id=None,
                    offsetId=None,
                    displayName=None,
                ),
            )
        )
    state, stored = await subject.build_n_labware_with_ids(
        pool_primary_definition=sentinel.primary_labware_def,
        pool_adapter_definition=None,
        pool_lid_definition=sentinel.lid_labware_def,
        module_id="module-id",
        ids=ids,
        current_contained_labware=current_contains,
        equipment=equipment,
    )
    assert stored == current_contains + ids
    location_ids: dict[str, LabwareLocation] = {}
    ids_flat: list[str] = []
    defs: dict[str, LabwareDefinition] = {}
    for group in ids:
        assert group.lidLabwareId
        location_ids[group.primaryLabwareId] = InStackerHopperLocation(
            moduleId="module-id"
        )
        location_ids[group.lidLabwareId] = OnLabwareLocation(
            labwareId=group.primaryLabwareId
        )
        ids_flat.extend([group.primaryLabwareId, group.lidLabwareId])
        defs[group.primaryLabwareId] = sentinel.primary_labware_def
        defs[group.lidLabwareId] = sentinel.lid_labware_def

    assert state == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id="module-id",
            contained_labware_bottom_first=current_contains + ids,
        ),
        batch_loaded_labware=BatchLoadedLabwareUpdate(
            new_locations_by_id=location_ids,
            offset_ids_by_id={id: None for id in ids_flat},
            display_names_by_id={id: None for id in ids_flat},
            definitions_by_id=defs,
        ),
        labware_lid=LabwareLidUpdate(
            parent_labware_ids=[id_group.primaryLabwareId for id_group in ids],
            lid_ids=[id_group.lidLabwareId for id_group in ids],
        ),
    )


@pytest.mark.parametrize(
    "ids,current_contains",
    [
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId="lid-id-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId="adapter-id-2",
                    lidLabwareId="lid-id-2",
                ),
            ],
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-1",
                    adapterLabwareId="adapter-pre-1",
                    lidLabwareId="lid-pre-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-pre-2",
                    adapterLabwareId="adapter-pre-2",
                    lidLabwareId="lid-pre-2",
                ),
            ],
            id="has-contents",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId="lid-id-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-2",
                    adapterLabwareId="adapter-id-2",
                    lidLabwareId="lid-id-2",
                ),
            ],
            [],
            id="no-contents",
        ),
    ],
)
async def test_build_n_labware_happypath_primary_and_lid_and_adapter(
    ids: list[StackerStoredLabwareGroup],
    current_contains: list[StackerStoredLabwareGroup],
    equipment: EquipmentHandler,
    decoy: Decoy,
) -> None:
    """It should build the specified labware."""
    for id_group in ids:
        assert id_group.lidLabwareId
        assert id_group.adapterLabwareId
        decoy.when(
            await equipment.load_labware_pool_from_definitions(
                pool_primary_definition=sentinel.primary_labware_def,
                pool_adapter_definition=sentinel.adapter_labware_def,
                pool_lid_definition=sentinel.lid_labware_def,
                location=InStackerHopperLocation(moduleId="module-id"),
                primary_id=id_group.primaryLabwareId,
                adapter_id=id_group.adapterLabwareId,
                lid_id=id_group.lidLabwareId,
            )
        ).then_return(
            LoadedLabwarePoolData(
                primary_labware=LoadedLabware(
                    id=id_group.primaryLabwareId,
                    loadName="some-load-name",
                    definitionUri="some-uri",
                    location=OnLabwareLocation(labwareId=id_group.adapterLabwareId),
                    lid_id=id_group.lidLabwareId,
                    offsetId=None,
                    displayName=None,
                ),
                adapter_labware=LoadedLabware(
                    id=id_group.adapterLabwareId,
                    loadName="adapter-load-name",
                    definitionUri="adapter-uri",
                    location=InStackerHopperLocation(
                        moduleId="module-id",
                    ),
                    lid_id=None,
                    offsetId=None,
                    displayName=None,
                ),
                lid_labware=LoadedLabware(
                    id=id_group.lidLabwareId,
                    loadName="lid-load-name",
                    definitionUri="lid-uri",
                    location=OnLabwareLocation(
                        labwareId=id_group.primaryLabwareId,
                    ),
                    lid_id=None,
                    offsetId=None,
                    displayName=None,
                ),
            )
        )
    state, stored = await subject.build_n_labware_with_ids(
        pool_primary_definition=sentinel.primary_labware_def,
        pool_adapter_definition=sentinel.adapter_labware_def,
        pool_lid_definition=sentinel.lid_labware_def,
        module_id="module-id",
        ids=ids,
        current_contained_labware=current_contains,
        equipment=equipment,
    )
    assert stored == current_contains + ids
    location_ids: dict[str, LabwareLocation] = {}
    ids_flat: list[str] = []
    defs: dict[str, LabwareDefinition] = {}
    for group in ids:
        assert group.lidLabwareId
        assert group.adapterLabwareId
        location_ids[group.primaryLabwareId] = OnLabwareLocation(
            labwareId=group.adapterLabwareId
        )
        location_ids[group.adapterLabwareId] = InStackerHopperLocation(
            moduleId="module-id"
        )
        location_ids[group.lidLabwareId] = OnLabwareLocation(
            labwareId=group.primaryLabwareId
        )
        ids_flat.extend(
            [group.primaryLabwareId, group.lidLabwareId, group.adapterLabwareId]
        )
        defs[group.primaryLabwareId] = sentinel.primary_labware_def
        defs[group.lidLabwareId] = sentinel.lid_labware_def
        defs[group.adapterLabwareId] = sentinel.adapter_labware_def

    assert state == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id="module-id",
            contained_labware_bottom_first=current_contains + ids,
        ),
        batch_loaded_labware=BatchLoadedLabwareUpdate(
            new_locations_by_id=location_ids,
            offset_ids_by_id={id: None for id in ids_flat},
            display_names_by_id={id: None for id in ids_flat},
            definitions_by_id=defs,
        ),
        labware_lid=LabwareLidUpdate(
            parent_labware_ids=[id_group.primaryLabwareId for id_group in ids],
            lid_ids=[id_group.lidLabwareId for id_group in ids],
        ),
    )


@pytest.mark.parametrize(
    "initial_labware,initial_count,current_count,results",
    [
        pytest.param(
            None,
            3,
            0,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="generated-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="generated-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="generated-3",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            id="generate-all-from-empty",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-3",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            None,
            0,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-3",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            id="specify-all-from-empty",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            None,
            2,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            id="specify-all-from-nonempty",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-3",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            3,
            0,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-3",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            id="specify-all-with-count-from-empty",
        ),
        pytest.param(
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            1,
            2,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="specified-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            id="specify-all-with-count-from-nonempty",
        ),
    ],
)
def test_build_ids_to_fill_variants(
    initial_labware: list[StackerStoredLabwareGroup] | None,
    initial_count: int | None,
    current_count: int,
    results: list[StackerStoredLabwareGroup],
    model_utils: ModelUtils,
    decoy: Decoy,
) -> None:
    """It should build ids appropriately."""
    max_count = 3
    generated_id_counts = 0

    def _new_id() -> str:
        nonlocal generated_id_counts
        generated_id_counts += 1
        return f"generated-{generated_id_counts}"

    decoy.when(model_utils.generate_id()).then_do(_new_id)

    assert (
        subject.build_ids_to_fill(
            False,
            False,
            initial_labware,
            initial_count,
            max_count,
            current_count,
            model_utils,
        )
        == results
    )


@pytest.mark.parametrize("using_list", [True, False])
@pytest.mark.parametrize(
    "specified_count,current_count",
    [
        pytest.param(
            3,
            0,
            id="empty",
        ),
        pytest.param(
            2,
            1,
            id="nonempty",
        ),
    ],
)
def test_build_ids_to_fill_fails_on_too_many_specified(
    using_list: bool,
    specified_count: int,
    current_count: int,
    model_utils: ModelUtils,
) -> None:
    """It should prevent you from specifying too many labware."""
    if using_list:
        specified_labware_list = [
            StackerStoredLabwareGroup(
                primaryLabwareId=f"primary-id-{idx}",
                adapterLabwareId=None,
                lidLabwareId=None,
            )
            for idx in range(specified_count)
        ]
        specified_labware_count = None
    else:
        specified_labware_list = None
        specified_labware_count = specified_count

    with pytest.raises(
        CommandPreconditionViolated,
        match=".*were requested to be stored, but the stacker can hold only.*",
    ):
        subject.build_ids_to_fill(
            False,
            False,
            specified_labware_list,
            specified_labware_count,
            2,
            current_count,
            model_utils,
        )


def test_build_ids_to_fill_fails_on_already_at_capacity(
    model_utils: ModelUtils,
) -> None:
    """If nothing is specified but already at max, it should fail."""
    with pytest.raises(
        CommandPreconditionViolated,
        match=".*already full.*",
    ):
        subject.build_ids_to_fill(
            False,
            False,
            None,
            None,
            2,
            2,
            model_utils,
        )


def test_build_ids_to_fill_fails_on_mismatched_count(model_utils: ModelUtils) -> None:
    """It should prevent you from specifying too many labware."""
    with pytest.raises(
        CommandPreconditionViolated,
        match="If initialCount and initialStoredLabware are both specified,.*",
    ):
        subject.build_ids_to_fill(
            False,
            False,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="primary-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            2,
            2,
            0,
            model_utils,
        )


@pytest.mark.parametrize("has_adapter", [True, False])
@pytest.mark.parametrize("has_lid", [True, False])
def test_build_ids_to_fill_builds_specified_components(
    has_adapter: bool, has_lid: bool, model_utils: ModelUtils, decoy: Decoy
) -> None:
    """It should build adapter and labware ids only where necessary."""
    generated_id_counts = 0

    def _new_id() -> str:
        nonlocal generated_id_counts
        generated_id_counts += 1
        return f"generated-{generated_id_counts}"

    decoy.when(model_utils.generate_id()).then_do(_new_id)
    subject.build_ids_to_fill(
        has_adapter,
        has_lid,
        None,
        1,
        2,
        0,
        model_utils,
    ) == [
        StackerStoredLabwareGroup(
            primaryLabwareId="generated-1",
            adapterLabwareId="generated-2" if has_adapter else None,
            lidLabwareId=(
                "generated-3"
                if has_adapter and has_lid
                else ("generated-2" if has_lid else None)
            ),
        )
    ]


@pytest.mark.parametrize(
    "group,locations,offsets",
    [
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-id", adapterLabwareId=None, lidLabwareId=None
            ),
            {"primary-id": ModuleLocation(moduleId="stacker-id")},
            {"primary-id": "primary-base-id"},
            id="primary-only",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-id",
                adapterLabwareId="adapter-id",
                lidLabwareId=None,
            ),
            {
                "adapter-id": ModuleLocation(moduleId="stacker-id"),
                "primary-id": OnLabwareLocation(labwareId="adapter-id"),
            },
            {
                "primary-id": "primary-on-adapter-base-id",
                "adapter-id": "adapter-base-id",
            },
            id="primary-and-adapter",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-id",
                adapterLabwareId=None,
                lidLabwareId="lid-id",
            ),
            {
                "primary-id": ModuleLocation(moduleId="stacker-id"),
                "lid-id": OnLabwareLocation(labwareId="primary-id"),
            },
            {"primary-id": "primary-base-id", "lid-id": "lid-primary-base-id"},
            id="primary-and-lid",
        ),
        pytest.param(
            StackerStoredLabwareGroup(
                primaryLabwareId="primary-id",
                adapterLabwareId="adapter-id",
                lidLabwareId="lid-id",
            ),
            {
                "adapter-id": ModuleLocation(moduleId="stacker-id"),
                "primary-id": OnLabwareLocation(labwareId="adapter-id"),
                "lid-id": OnLabwareLocation(labwareId="primary-id"),
            },
            {
                "primary-id": "primary-on-adapter-base-id",
                "adapter-id": "adapter-base-id",
                "lid-id": "lid-on-primary-on-adapter-base-id",
            },
            id="primary-adapter-and-lid",
        ),
    ],
)
def test_build_retrieve_labware_move_updates(
    group: StackerStoredLabwareGroup,
    locations: dict[str, LabwareLocation],
    offsets: dict[str, None],
    state_view: StateView,
    decoy: Decoy,
) -> None:
    """It should build appropriate data for batch labware location."""
    stacker = FlexStackerSubState(
        module_id=FlexStackerId("stacker-id"),
        pool_primary_definition=sentinel.pool_primary_definition,
        pool_adapter_definition=sentinel.pool_adapter_definition,
        pool_lid_definition=sentinel.pool_lid_definition,
        max_pool_count=3,
        pool_overlap=1,
        pool_height=0,
        contained_labware_bottom_first=[group],
    )
    decoy.when(
        state_view.geometry.get_projected_offset_location(
            ModuleLocation(moduleId="stacker-id")
        )
    ).then_return([sentinel.offset_location_base])
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "adapter-labware-uri", [sentinel.offset_location_base]
        )
    ).then_return(
        LabwareOffset.model_construct(id="adapter-base-id")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "primary-labware-uri", [sentinel.offset_location_base]
        )
    ).then_return(
        LabwareOffset.model_construct(id="primary-base-id")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "primary-labware-uri",
            [
                OnLabwareOffsetLocationSequenceComponent.model_construct(
                    labwareUri="adapter-labware-uri",
                ),
                sentinel.offset_location_base,
            ],
        )
    ).then_return(
        LabwareOffset.model_construct(id="primary-on-adapter-base-id")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "lid-labware-uri",
            [
                OnLabwareOffsetLocationSequenceComponent.model_construct(
                    labwareUri="primary-labware-uri"
                ),
                sentinel.offset_location_base,
            ],
        )
    ).then_return(
        LabwareOffset.model_construct(id="lid-primary-base-id")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "lid-labware-uri",
            [
                OnLabwareOffsetLocationSequenceComponent.model_construct(
                    labwareUri="primary-labware-uri"
                ),
                OnLabwareOffsetLocationSequenceComponent.model_construct(
                    labwareUri="adapter-labware-uri"
                ),
                sentinel.offset_location_base,
            ],
        )
    ).then_return(
        LabwareOffset.model_construct(id="lid-on-primary-on-adapter-base-id")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.pool_primary_definition)
    ).then_return(LabwareUri("primary-labware-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.pool_adapter_definition)
    ).then_return(LabwareUri("adapter-labware-uri"))
    decoy.when(
        state_view.labware.get_uri_from_definition(sentinel.pool_lid_definition)
    ).then_return(LabwareUri("lid-labware-uri"))
    check_locations, check_offsets = subject.build_retrieve_labware_move_updates(
        group, stacker, state_view
    )
    assert check_locations == locations
    assert check_offsets == offsets



import platform
from typing import List, Tuple

import pytest
from decoy import Decoy
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parameters.csv_parameter_interface import CSVParameter


@pytest.fixture
def api_version() -> APIVersion:
    """The API version under test."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def csv_file_basic() -> bytes:
    """A basic CSV file with quotes around strings."""
    return b'"x","y","z"\n"a",1,2\n"b",3,4\n"c",5,6'


@pytest.fixture
def csv_file_no_quotes() -> bytes:
    """A basic CSV file with no quotes around strings."""
    return b"x,y,z\na,1,2\nb,3,4\nc,5,6"


@pytest.fixture
def csv_file_preceding_spaces() -> bytes:
    """A basic CSV file with quotes around strings and spaces preceding non-initial columns."""
    return b'"x", "y", "z"\n"a", 1, 2\n"b", 3, 4\n"c", 5, 6'


@pytest.fixture
def csv_file_mixed_quotes() -> bytes:
    """A basic CSV file with both string quotes and escaped quotes."""
    return b'head,er\n"a,b,c",def\n"""ghi""","jkl"'


@pytest.fixture
def csv_file_different_delimiter() -> bytes:
    """A basic CSV file with a non-comma delimiter."""
    return b"x:y:z\na,:1,:2\nb,:3,:4\nc,:5,:6"


@pytest.fixture()
def csv_file_long() -> bytes:
    """A long CSV file from a customer that caused the sniffer to fail when it only looked at the first 1024 bytes."""
    return b"""
Source Labware,Source Slot,Source Well,Source Height,Dest Labware,Dest Slot,Dest Well,Volume
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical,0,A1,0,opentrons_24_aluminumblock_nest_0.5ml_screwcap,0,A1,100
""".strip()


@pytest.fixture
def csv_file_basic_trailing_empty() -> Tuple[bytes, List[List[str]]]:
    """A basic CSV file with quotes around strings and a trailing newline."""
    return (
        b'"x","y","z"\n"a",1,2\n"b",3,4\n"c",5,6\n',
        [["x", "y", "z"], ["a", "1", "2"], ["b", "3", "4"], ["c", "5", "6"]],
    )


@pytest.fixture
def csv_file_basic_three_trailing_empty() -> Tuple[bytes, List[List[str]]]:
    """A basic CSV file with quotes around strings and three trailing newlines."""
    return (
        b'"x","y","z"\n"a",1,2\n"b",3,4\n"c",5,6\n\n\n',
        [["x", "y", "z"], ["a", "1", "2"], ["b", "3", "4"], ["c", "5", "6"]],
    )


@pytest.fixture
def csv_file_empty_row_and_trailing_empty() -> Tuple[bytes, List[List[str]]]:
    """A basic CSV file with quotes around strings, an empty row, and a trailing newline."""
    return (
        b'"x","y","z"\n\n"b",3,4\n"c",5,6\n',
        [["x", "y", "z"], [], ["b", "3", "4"], ["c", "5", "6"]],
    )


@pytest.fixture
def csv_file_windows_empty_row_trailing_empty() -> Tuple[bytes, List[List[str]]]:
    """A basic CSV file with quotes around strings, Windows-style newlines, an empty row, and a trailing newline."""
    return (
        b'"x","y","z"\r\n\r\n"b",3,4\r\n"c",5,6\r\n',
        [["x", "y", "z"], [], ["b", "3", "4"], ["c", "5", "6"]],
    )


def test_csv_parameter(
    decoy: Decoy, api_version: APIVersion, csv_file_basic: bytes
) -> None:
    """It should load the CSV parameter and provide access to the file, contents, and rows."""
    subject = CSVParameter(csv_file_basic, api_version)

    # On Windows, you can't open a NamedTemporaryFile a second time, which breaks the code under test.
    # Because of the way CSV analysis works this code will only ever be run on the actual OT-2/Flex hardware,
    # so we skip testing and instead assert that we get a PermissionError on Windows (to ensure this
    # test gets fixed in case we ever refactor the file opening.)
    if platform.system() != "Windows":
        assert subject.file.readable()
        assert not subject.file.writable()
        assert subject.file.read() == '"x","y","z"\n"a",1,2\n"b",3,4\n"c",5,6'
    else:
        with pytest.raises(PermissionError):
            subject.file
    assert subject.contents == '"x","y","z"\n"a",1,2\n"b",3,4\n"c",5,6'
    assert subject.parse_as_csv()[0] == ["x", "y", "z"]


def test_csv_parameter_long_file(
    decoy: Decoy, api_version: APIVersion, csv_file_long: bytes
) -> None:
    """It should detect the CSV dialect for files of unlimited length."""
    # The previous implementation of parse_as_csv() passed only the first 1024 bytes of
    # the CSV file to the dialect sniffer, chopping up a line an unfortunate position,
    # causing the sniffer to fail with "Could not determine delimiter".
    subject = CSVParameter(csv_file_long, api_version)
    parsed_rows = subject.parse_as_csv()
    assert len(parsed_rows) == 10
    assert len(parsed_rows[0]) == 8


@pytest.mark.parametrize(
    "csv_file",
    [
        lazy_fixture("csv_file_basic"),
        lazy_fixture("csv_file_no_quotes"),
        lazy_fixture("csv_file_preceding_spaces"),
    ],
)
def test_csv_parameter_rows(
    decoy: Decoy,
    api_version: APIVersion,
    csv_file: bytes,
) -> None:
    """It should load the rows as all strings even with no quotes or leading spaces."""
    subject = CSVParameter(csv_file, api_version)

    assert len(subject.parse_as_csv()) == 4
    assert subject.parse_as_csv()[0] == ["x", "y", "z"]
    assert subject.parse_as_csv()[1] == ["a", "1", "2"]


def test_csv_parameter_mixed_quotes(
    decoy: Decoy,
    api_version: APIVersion,
    csv_file_mixed_quotes: bytes,
) -> None:
    """It should load the rows with no quotes, quotes and escaped quotes with double quotes."""
    subject = CSVParameter(csv_file_mixed_quotes, api_version)

    assert len(subject.parse_as_csv()) == 3
    assert subject.parse_as_csv()[0] == ["head", "er"]
    assert subject.parse_as_csv()[1] == ["a,b,c", "def"]
    assert subject.parse_as_csv()[2] == ['"ghi"', "jkl"]


def test_csv_parameter_additional_kwargs(
    decoy: Decoy,
    api_version: APIVersion,
    csv_file_different_delimiter: bytes,
) -> None:
    """It should load the rows with a different delimiter."""
    subject = CSVParameter(csv_file_different_delimiter, api_version)
    rows = subject.parse_as_csv(delimiter=":")

    assert len(rows) == 4
    assert rows[0] == ["x", "y", "z"]
    assert rows[1] == ["a,", "1,", "2"]


def test_csv_parameter_dont_detect_dialect(
    decoy: Decoy,
    api_version: APIVersion,
    csv_file_preceding_spaces: bytes,
) -> None:
    """It should load the rows without trying to detect the dialect."""
    subject = CSVParameter(csv_file_preceding_spaces, api_version)
    rows = subject.parse_as_csv(detect_dialect=False)

    assert rows[0] == ["x", ' "y"', ' "z"']
    assert rows[1] == ["a", " 1", " 2"]


@pytest.mark.parametrize(
    "csv_file_fixture",
    [
        "csv_file_basic_trailing_empty",
        "csv_file_basic_three_trailing_empty",
        "csv_file_empty_row_and_trailing_empty",
        "csv_file_windows_empty_row_trailing_empty",
    ],
)
def test_csv_parameter_trailing_empties(
    decoy: Decoy,
    api_version: APIVersion,
    request: pytest.FixtureRequest,
    csv_file_fixture: str,
) -> None:
    """It should load the rows as all strings.  Empty rows are allowed in the middle of the data but all trailing empty rows are removed."""
    # Get the fixture value
    csv_file: bytes
    expected_output: List[List[str]]
    csv_file, expected_output = request.getfixturevalue(csv_file_fixture)

    subject = CSVParameter(csv_file, api_version)
    parsed_csv = subject.parse_as_csv()

    assert parsed_csv == expected_output, (
        f"Expected {expected_output}, but got {parsed_csv}"
    )
    assert len(parsed_csv) == len(expected_output), (
        f"Expected {len(expected_output)} rows, but got {len(parsed_csv)}"
    )



"""Test blow-out command."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    PipetteOverpressureError,
    StallOrCollisionDetectedError,
)

from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine import (
    DeckPoint,
    WellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.protocol_engine.commands import (
    BlowOutImplementation,
    BlowOutParams,
    BlowOutResult,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.commands.pipetting_common import OverpressureError
from opentrons.protocol_engine.execution import (
    MovementHandler,
    PipettingHandler,
)
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId
from opentrons.types import Point


@pytest.fixture
def subject(
    state_view: StateView,
    hardware_api: HardwareControlAPI,
    movement: MovementHandler,
    model_utils: ModelUtils,
    pipetting: PipettingHandler,
) -> BlowOutImplementation:
    """Get the impelementation subject."""
    return BlowOutImplementation(
        state_view=state_view,
        movement=movement,
        hardware_api=hardware_api,
        pipetting=pipetting,
        model_utils=model_utils,
    )


async def test_blow_out_implementation(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: BlowOutImplementation,
) -> None:
    """Test BlowOut command execution."""
    location = WellLocation(origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1))

    data = BlowOutParams(
        pipetteId="pipette-id",
        labwareId="labware-id",
        wellName="C6",
        wellLocation=location,
        flowRate=1.234,
    )

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="C6",
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=1, y=2, z=3))

    result = await subject.execute(data)

    assert result == SuccessData(
        public=BlowOutResult(position=DeckPoint(x=1, y=2, z=3)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id",
                    well_name="C6",
                ),
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            ),
            pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                pipette_id="pipette-id", clean_tip=False
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="pipette-id", ready_to_aspirate=False
            ),
        ),
    )

    decoy.verify(
        await pipetting.blow_out_in_place(pipette_id="pipette-id", flow_rate=1.234),
        times=1,
    )


async def test_overpressure_error(
    decoy: Decoy,
    pipetting: PipettingHandler,
    subject: BlowOutImplementation,
    model_utils: ModelUtils,
    movement: MovementHandler,
) -> None:
    """It should return an overpressure error if the hardware API indicates that."""
    pipette_id = "pipette-id"

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    location = WellLocation(origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1))

    data = BlowOutParams(
        pipetteId="pipette-id",
        labwareId="labware-id",
        wellName="C6",
        wellLocation=location,
        flowRate=1.234,
    )

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        True
    )

    decoy.when(
        await pipetting.blow_out_in_place(pipette_id="pipette-id", flow_rate=1.234)  # type: ignore[func-returns-value]
    ).then_raise(PipetteOverpressureError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)
    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="C6",
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=1, y=2, z=3))

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=OverpressureError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={"retryLocation": (1, 2, 3)},
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id",
                    well_name="C6",
                ),
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="pipette-id"
            ),
        ),
        state_update_if_false_positive=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id",
                    well_name="C6",
                ),
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            ),
        ),
    )


async def test_stall_error(
    decoy: Decoy,
    pipetting: PipettingHandler,
    subject: BlowOutImplementation,
    model_utils: ModelUtils,
    movement: MovementHandler,
) -> None:
    """It should return an overpressure error if the hardware API indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "C6"
    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    location = WellLocation(origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1))

    data = BlowOutParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=location,
        flowRate=1.234,
    )

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)
    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_raise(StallOrCollisionDetectedError())

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
        ),
    )



from pathlib import Path
from typing import Callable, Union

import pytest
from mock import patch
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.errors.exceptions import InvalidLiquidClassName
from opentrons_shared_data.pipette import (
    load_data as load_pipette_data,
)
from opentrons_shared_data.pipette import (
    mutable_configurations,
    pipette_definition,
)
from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
)
from opentrons_shared_data.pipette import (
    types as pip_types,
)
from opentrons_shared_data.pipette.types import PipetteModel

from opentrons.calibration_storage import types as cal_types
from opentrons.hardware_control import types
from opentrons.hardware_control.instruments.ot2 import (
    instrument_calibration,
)
from opentrons.hardware_control.instruments.ot2 import (
    pipette as ot2_pipette,
)
from opentrons.hardware_control.instruments.ot3 import (
    instrument_calibration as ot3_calibration,
)
from opentrons.hardware_control.instruments.ot3 import (
    pipette as ot3_pipette,
)
from opentrons.types import Mount, Point

OT2_PIP_CAL = instrument_calibration.PipetteOffsetByPipetteMount(
    offset=Point(0, 0, 0),
    source=cal_types.SourceType.user,
    status=cal_types.CalibrationStatus(),
)

OT3_PIP_CAL = ot3_calibration.PipetteOffsetByPipetteMount(
    offset=Point(0, 0, 0),
    source=cal_types.SourceType.user,
    status=cal_types.CalibrationStatus(),
)


@pytest.fixture
def hardware_pipette_ot2() -> Callable[
    [PipetteModel, instrument_calibration.PipetteOffsetByPipetteMount, str],
    ot2_pipette.Pipette,
]:
    def _create_pipette(
        model: PipetteModel,
        calibration: instrument_calibration.PipetteOffsetByPipetteMount = OT2_PIP_CAL,
        id: str = "testID",
    ) -> ot2_pipette.Pipette:
        pipette_model = pipette_load_name.convert_pipette_model(model)
        configurations = mutable_configurations.load_with_mutable_configurations(
            pipette_model, Path("fake/path"), "testiId"
        )
        return ot2_pipette.Pipette(configurations, calibration, id)

    return _create_pipette


@pytest.fixture
def hardware_pipette_ot3() -> Callable[
    [
        pipette_definition.PipetteModelVersionType,
        ot3_calibration.PipetteOffsetByPipetteMount,
        str,
    ],
    ot3_pipette.Pipette,
]:
    def _create_pipette(
        model: pipette_definition.PipetteModelVersionType,
        calibration: ot3_calibration.PipetteOffsetByPipetteMount = OT3_PIP_CAL,
        id: str = "testID",
    ) -> ot3_pipette.Pipette:
        return ot3_pipette.Pipette(
            load_pipette_data.load_definition(
                model.pipette_type,
                model.pipette_channels,
                model.pipette_version,
                model.oem_type,
            ),
            calibration,
            id,
        )

    return _create_pipette


@pytest.mark.parametrize(
    argnames=["pipette_builder", "model"],
    argvalues=[
        [lazy_fixture("hardware_pipette_ot2"), "p10_single_v1"],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v1.0")),
        ],
    ],
)
def test_tip_tracking(
    pipette_builder: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
    model: Union[str, pipette_definition.PipetteModelVersionType],
) -> None:
    hw_pipette = pipette_builder(model)
    hw_pipette.remove_tip()
    assert not hw_pipette.has_tip
    tip_length = 25.0
    hw_pipette.add_tip(tip_length)
    assert hw_pipette.has_tip
    with pytest.raises(AssertionError):
        hw_pipette.add_tip(tip_length)
    hw_pipette.remove_tip()
    assert not hw_pipette.has_tip
    hw_pipette.remove_tip()


@pytest.mark.parametrize(
    argnames=["pipette_builder", "model", "nozzle_offset"],
    argvalues=[
        [lazy_fixture("hardware_pipette_ot2"), "p10_single_v1", Point(0, 0, 12.0)],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v3.3")),
            Point(-8.0, -22.0, -259.15),
        ],
    ],
)
def test_tip_nozzle_position_tracking(
    pipette_builder: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
    model: Union[str, pipette_definition.PipetteModelVersionType],
    nozzle_offset: Point,
) -> None:
    hw_pipette = pipette_builder(model)
    # default pipette offset is[0, 0, 0], only nozzle offset would be used
    # to determine critical point
    assert hw_pipette.critical_point() == nozzle_offset
    assert hw_pipette.critical_point(types.CriticalPoint.NOZZLE) == nozzle_offset
    assert hw_pipette.critical_point(types.CriticalPoint.TIP) == nozzle_offset
    tip_length = 25.0
    hw_pipette.add_tip(tip_length)
    new = nozzle_offset._replace(z=nozzle_offset.z - tip_length)
    assert hw_pipette.critical_point() == new
    assert hw_pipette.critical_point(types.CriticalPoint.NOZZLE) == nozzle_offset
    assert hw_pipette.critical_point(types.CriticalPoint.TIP) == new
    hw_pipette.remove_tip()
    assert hw_pipette.critical_point() == nozzle_offset
    assert hw_pipette.critical_point(types.CriticalPoint.NOZZLE) == nozzle_offset
    assert hw_pipette.critical_point(types.CriticalPoint.TIP) == nozzle_offset


@pytest.mark.parametrize(
    argnames=["pipette_builder", "model", "calibration"],
    argvalues=[
        [
            lazy_fixture("hardware_pipette_ot2"),
            "p10_single_v1",
            instrument_calibration.PipetteOffsetByPipetteMount(
                offset=Point(10, 10, 10),
                source=cal_types.SourceType.user,
                status=cal_types.CalibrationStatus(),
            ),
        ],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v1.0")),
            ot3_calibration.PipetteOffsetByPipetteMount(
                offset=Point(10, 10, 10),
                source=cal_types.SourceType.user,
                status=cal_types.CalibrationStatus(),
            ),
        ],
    ],
)
def test_critical_points_pipette_offset(
    pipette_builder: Callable[
        [
            Union[str, pipette_definition.PipetteModelVersionType],
            Union[
                instrument_calibration.PipetteOffsetByPipetteMount,
                ot3_calibration.PipetteOffsetByPipetteMount,
            ],
        ],
        ot3_pipette.Pipette,
    ],
    model: Union[str, pipette_definition.PipetteModelVersionType],
    calibration: Union[
        instrument_calibration.PipetteOffsetByPipetteMount,
        ot3_calibration.PipetteOffsetByPipetteMount,
    ],
) -> None:
    hw_pipette = pipette_builder(model, calibration)
    # pipette offset + nozzle offset to determine critical point
    offsets = calibration.offset + hw_pipette.nozzle_offset
    assert hw_pipette.critical_point() == offsets
    assert hw_pipette.critical_point(types.CriticalPoint.NOZZLE) == offsets
    assert hw_pipette.critical_point(types.CriticalPoint.TIP) == offsets
    tip_length = 25.0
    hw_pipette.add_tip(tip_length)
    new = offsets._replace(z=offsets.z - tip_length)
    assert hw_pipette.critical_point() == new
    assert hw_pipette.critical_point(types.CriticalPoint.NOZZLE) == offsets
    assert hw_pipette.critical_point(types.CriticalPoint.TIP) == new
    hw_pipette.remove_tip()
    assert hw_pipette.critical_point() == offsets
    assert hw_pipette.critical_point(types.CriticalPoint.NOZZLE) == offsets
    assert hw_pipette.critical_point(types.CriticalPoint.TIP) == offsets


@pytest.mark.parametrize(
    argnames=["pipette_builder", "model", "max_volume"],
    argvalues=[
        [lazy_fixture("hardware_pipette_ot2"), "p10_single_v1", 10.0],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v1.0")),
            1000.0,
        ],
    ],
)
def test_volume_tracking(
    pipette_builder: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
    model: Union[str, pipette_definition.PipetteModelVersionType],
    max_volume: float,
) -> None:
    hw_pipette = pipette_builder(model)
    assert hw_pipette.current_volume == 0.0
    assert hw_pipette.available_volume == max_volume
    assert hw_pipette.ok_to_add_volume(max_volume - 0.1)
    hw_pipette.set_current_volume(0.1)
    with pytest.raises(AssertionError):
        hw_pipette.set_current_volume(max_volume + 0.1)
    with pytest.raises(AssertionError):
        hw_pipette.set_current_volume(-1)
    assert hw_pipette.current_volume == 0.1
    hw_pipette.remove_current_volume(0.1)
    with pytest.raises(AssertionError):
        hw_pipette.remove_current_volume(0.1)
    assert hw_pipette.current_volume == 0.0
    hw_pipette.set_current_volume(max_volume)
    assert not hw_pipette.ok_to_add_volume(0.1)
    with pytest.raises(AssertionError):
        hw_pipette.add_current_volume(0.1)
    assert hw_pipette.current_volume == max_volume


@pytest.mark.ot2_only
def test_config_update(
    hardware_pipette_ot2: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot2_pipette.Pipette
    ],
) -> None:
    hw_pipette = hardware_pipette_ot2("p10_single_v1")
    sample_plunger_pos = {"top": 19.5}
    hw_pipette.update_config_item(sample_plunger_pos)
    assert hw_pipette.plunger_positions.top == sample_plunger_pos.get("top")


@pytest.mark.ot2_only
def test_flow_rate_setting(
    hardware_pipette_ot2: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot2_pipette.Pipette
    ],
) -> None:
    hw_pipette = hardware_pipette_ot2("p10_single_v1")
    # pipettes should load settings from config at init time
    assert hw_pipette.aspirate_flow_rate == hw_pipette.aspirate_flow_rates_lookup["2.0"]
    assert hw_pipette.dispense_flow_rate == hw_pipette.dispense_flow_rates_lookup["2.0"]
    assert hw_pipette.blow_out_flow_rate == hw_pipette.blow_out_flow_rates_lookup["2.0"]
    # changing flow rates with normal property access shouldn't touch
    # config or other flow rates
    hw_pipette.aspirate_flow_rate = 2
    assert hw_pipette.aspirate_flow_rate == 2
    assert hw_pipette.dispense_flow_rate == hw_pipette.dispense_flow_rates_lookup["2.0"]
    assert hw_pipette.blow_out_flow_rate == hw_pipette.blow_out_flow_rates_lookup["2.0"]
    hw_pipette.dispense_flow_rate = 3
    assert hw_pipette.aspirate_flow_rate == 2
    assert hw_pipette.dispense_flow_rate == 3
    assert hw_pipette.blow_out_flow_rate == hw_pipette.blow_out_flow_rates_lookup["2.0"]
    hw_pipette.blow_out_flow_rate = 4
    assert hw_pipette.aspirate_flow_rate == 2
    assert hw_pipette.dispense_flow_rate == 3
    assert hw_pipette.blow_out_flow_rate == 4


@pytest.mark.parametrize(
    argnames=[
        "pipette_builder",
        "model",
        "expected_xy_critical_point",
        "expected_front_critical_point",
    ],
    argvalues=[
        [
            lazy_fixture("hardware_pipette_ot2"),
            "p10_single_v1",
            Point(0, 0, 12.0),
            Point(0, 0, 12.0),
        ],
        [
            lazy_fixture("hardware_pipette_ot2"),
            "p300_multi_v2.0",
            Point(0.0, 0.0, 35.52),
            Point(0.0, -31.5, 35.52),
        ],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v3.3")),
            Point(-8.0, -22.0, -259.15),
            Point(-8.0, -22.0, -259.15),
        ],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(PipetteModel("p1000_multi_v3.3")),
            Point(-8.0, -47.5, -259.15),
            Point(-8.0, -79.0, -259.15),
        ],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(
                PipetteModel("p1000_96"), PipetteModel("3.3")
            ),
            Point(13.5, -57.0, -259.15),
            Point(-36.0, -88.5, -259.15),
        ],
    ],
)
def test_alternative_critical_points(
    pipette_builder: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
    model: Union[str, pipette_definition.PipetteModelVersionType],
    expected_xy_critical_point: Point,
    expected_front_critical_point: Point,
) -> None:
    hw_pipette = pipette_builder(model)
    assert (
        hw_pipette.critical_point(types.CriticalPoint.XY_CENTER)
        == expected_xy_critical_point
    )
    assert (
        hw_pipette.critical_point(types.CriticalPoint.FRONT_NOZZLE)
        == expected_front_critical_point
    )


@pytest.mark.parametrize(
    argnames=["pipette_builder", "model", "calibration"],
    argvalues=[
        [
            lazy_fixture("hardware_pipette_ot2"),
            "p10_single_v1",
            instrument_calibration.PipetteOffsetByPipetteMount(
                offset=Point(1, 1, 1),
                source=cal_types.SourceType.user,
                status=cal_types.CalibrationStatus(),
            ),
        ],
        [
            lazy_fixture("hardware_pipette_ot3"),
            pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v1.0")),
            ot3_calibration.PipetteOffsetByPipetteMount(
                offset=Point(1, 1, 1),
                source=cal_types.SourceType.user,
                status=cal_types.CalibrationStatus(),
            ),
        ],
    ],
)
def test_reset_instrument_offset(
    pipette_builder: Callable[
        [
            Union[str, pipette_definition.PipetteModelVersionType],
            Union[
                instrument_calibration.PipetteOffsetByPipetteMount,
                ot3_calibration.PipetteOffsetByPipetteMount,
            ],
        ],
        ot3_pipette.Pipette,
    ],
    model: Union[str, pipette_definition.PipetteModelVersionType],
    calibration: Union[
        instrument_calibration.PipetteOffsetByPipetteMount,
        ot3_calibration.PipetteOffsetByPipetteMount,
    ],
) -> None:
    hw_pipette = pipette_builder(model, calibration)
    assert hw_pipette.pipette_offset.offset == Point(1, 1, 1)
    hw_pipette.reset_pipette_offset(types.OT3Mount.LEFT, to_default=True)
    assert hw_pipette.pipette_offset.offset == Point(0, 0, 0)


def test_save_instrument_offset_ot2(
    hardware_pipette_ot2: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot2_pipette.Pipette
    ],
) -> None:
    # TODO (lc 10-31-2022) These tests would be much cleaner/easier to mock with
    # an InstrumentCalibrationProvider class (like robot calibration provider)
    # which should be done in a follow-up refactor.
    path_to_calibrations = "opentrons.hardware_control.instruments.ot2.pipette"
    hw_pipette = hardware_pipette_ot2("p10_single_v1")

    assert hw_pipette.pipette_offset.offset == Point(0, 0, 0)
    with patch(f"{path_to_calibrations}.load_pipette_offset") as load_cal:
        hw_pipette.save_pipette_offset(Mount.LEFT, Point(1.0, 2.0, 3.0))
        load_cal.assert_called_once_with("testID", Mount.LEFT)


def test_save_instrument_offset_ot3(
    hardware_pipette_ot3: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
) -> None:
    # TODO (lc 10-31-2022) These tests would be much cleaner/easier to mock with
    # an InstrumentCalibrationProvider class (like robot calibration provider)
    # which should be done in a follow-up refactor.
    path_to_calibrations = "opentrons.hardware_control.instruments.ot3.pipette"
    hw_pipette = hardware_pipette_ot3(
        pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v1.0"))
    )

    assert hw_pipette.pipette_offset.offset == Point(0, 0, 0)
    with (
        patch(f"{path_to_calibrations}.save_pipette_offset_calibration") as save_cal,
        patch(f"{path_to_calibrations}.load_pipette_offset") as load_cal,
    ):
        hw_pipette.save_pipette_offset(types.OT3Mount.LEFT, Point(1.0, 2.0, 3.0))

        save_cal.assert_called_once_with(
            "testID", types.OT3Mount.LEFT, Point(x=1.0, y=2.0, z=3.0)
        )
        load_cal.assert_called_once_with("testID", types.OT3Mount.LEFT)


def test_reload_instrument_cal_ot3(
    hardware_pipette_ot3: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
) -> None:
    old_pip = hardware_pipette_ot3(
        pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v1.0"))
    )
    # if only calibration is changed
    new_cal = ot3_calibration.PipetteOffsetByPipetteMount(
        offset=Point(3, 4, 5),
        source=cal_types.SourceType.user,
        status=cal_types.CalibrationStatus(),
    )
    new_pip, skipped = ot3_pipette._reload_and_check_skip(
        old_pip.config, old_pip, new_cal, use_old_aspiration_functions=False
    )

    assert skipped
    # it's the same pipette
    assert new_pip == old_pip


def test_tip_type_changing(
    hardware_pipette_ot3: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
) -> None:
    pip = hardware_pipette_ot3(
        pipette_load_name.convert_pipette_model(PipetteModel("p1000_single_v3.3"))
    )
    # at load, we have the max tip settings
    assert pip.working_volume == 1000
    # setting to the same thing works
    pip.set_tip_type_by_volume(1000)
    assert pip.working_volume == 1000
    # setting to other tip types works
    pip.set_tip_type_by_volume(200)
    assert pip.working_volume == 200

    # specifying a tip that doesn't exist raises an error
    with pytest.raises(InvalidLiquidClassName):
        pip.set_tip_type_by_volume(201)


def test_liquid_class_changing(
    hardware_pipette_ot3: Callable[
        [Union[str, pipette_definition.PipetteModelVersionType]], ot3_pipette.Pipette
    ],
) -> None:
    pip = hardware_pipette_ot3(
        pipette_load_name.convert_pipette_model(PipetteModel("p50_multi_v3.3"))
    )
    assert pip._liquid_class_name == pip_types.LiquidClasses.default
    pip.set_liquid_class_by_name("lowVolumeDefault")
    assert (
        pip._liquid_class_name.value == pip_types.LiquidClasses.lowVolumeDefault.value
    )
    with pytest.raises(InvalidLiquidClassName):
        pip.set_liquid_class_by_name("asdase")



from opentrons import containers, instruments

metadata = {
    "protocolName": "Testosaur",
    "author": "Opentrons <engineering@opentrons.com>",
    "description": 'A variant on "Dinosaur" for testing',
    "source": "Opentrons Repository",
}

p200rack = containers.load("tiprack-200ul", "5", "tiprack")  # type: ignore

# create a p200 pipette on robot axis B
p300 = instruments.P300_Single(mount="right", tip_racks=[p200rack])  # type: ignore

p300.pick_up_tip()

conts = [containers.load("96-PCR-flat", slot) for slot in ("8", "11")]  # type: ignore

# Uncomment these to test precision
# p300.move_to(robot.deck['11'])
# p300.move_to(robot.deck['6'])

for container in conts:
    p300.aspirate(10, container[0]).dispense(10, container[-1].top(5))

p300.drop_tip()
